// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    config::CpuConfig,
    instruction::{
        AddSubMOp, AluBranchMOp, AluCommonMOp, CommonMOp, LogicalMOp, MOpTrait, OutputIntegerMode,
        UnitOutRegNum, COMMON_MOP_SRC_LEN,
    },
    register::{FlagsMode, PRegFlagsPowerISA, PRegFlagsX86, PRegValue},
    unit::{
        unit_base::{unit_base, ExecuteEnd, ExecuteStart, UnitToRegAlloc},
        DynUnit, DynUnitWrapper, GlobalState, UnitKind, UnitMOp, UnitOutput, UnitResult,
        UnitResultCompleted, UnitTrait,
    },
};
use fayalite::{
    intern::{Intern, Interned},
    module::wire_with_loc,
    prelude::*,
    util::ready_valid::ReadyValid,
};
use std::{collections::HashMap, ops::RangeTo};

#[hdl]
fn add_sub<SrcCount: KnownSize>(
    mop: Expr<AddSubMOp<UnitOutRegNum<DynSize>, DynSize, SrcCount>>,
    pc: Expr<UInt<64>>,
    flags_mode: Expr<FlagsMode>,
    src_values: Expr<Array<PRegValue, { COMMON_MOP_SRC_LEN }>>,
) -> Expr<UnitResultCompleted<()>> {
    #[hdl]
    let AddSubMOp::<_, _, _> {
        alu_common,
        invert_src0,
        src1_is_carry_in,
        invert_carry_in,
        add_pc,
    } = mop;
    #[hdl]
    let AluCommonMOp::<_, _, _> {
        common,
        output_integer_mode,
    } = alu_common;
    let imm: Expr<UInt<64>> = CommonMOp::imm(common).cast_to_static();
    #[hdl]
    let carry_in_before_inversion = wire();
    connect(carry_in_before_inversion, false);
    #[hdl]
    let src1 = wire();
    connect(src1, 0u64);
    #[hdl]
    if src1_is_carry_in {
        #[hdl]
        match flags_mode {
            FlagsMode::PowerISA(_) => {
                connect(
                    carry_in_before_inversion,
                    PRegFlagsPowerISA::xer_ca(src_values[1].flags),
                );
            }
            FlagsMode::X86(_) => {
                connect(
                    carry_in_before_inversion,
                    PRegFlagsX86::cf(src_values[1].flags),
                );
            }
        }
    } else {
        connect(src1, src_values[1].int_fp);
    }
    #[hdl]
    let carry_in = wire();
    connect(carry_in, carry_in_before_inversion ^ invert_carry_in);
    #[hdl]
    let src0 = wire();
    connect(src0, src_values[0].int_fp);
    #[hdl]
    if invert_src0 {
        connect(src0, !src_values[0].int_fp);
    }
    #[hdl]
    let pc_or_zero = wire();
    connect(pc_or_zero, 0u64);
    #[hdl]
    if add_pc {
        connect(pc_or_zero, pc);
    }
    let sum_of_sliced = |slice: RangeTo<usize>| {
        src0[slice] + src1[slice] + src_values[2].int_fp[slice] + pc_or_zero[slice] + imm[slice]
    };
    #[hdl]
    let sum: UInt<64> = wire();
    connect(sum, sum_of_sliced(..64).cast_to_static());
    let carries =
        HashMap::<usize, _>::from_iter([4, 7, 8, 15, 16, 31, 32, 63, 64].into_iter().map(
            |bit_index| {
                let carry_at = wire_with_loc(
                    &format!("carry_at_{bit_index}"),
                    SourceLocation::caller(),
                    Bool,
                );
                connect(carry_at, sum_of_sliced(..bit_index)[bit_index]);
                (bit_index, carry_at)
            },
        ));

    #[hdl]
    let int_fp: UInt<64> = wire();

    #[hdl]
    let x86_cf = wire();
    #[hdl]
    let x86_af = wire();
    connect(x86_af, carries[&4]);
    #[hdl]
    let x86_of = wire();
    #[hdl]
    let x86_sf = wire();
    #[hdl]
    let x86_pf = wire();
    connect(x86_pf, sum[..8].parity_even());
    #[hdl]
    let x86_zf = wire();
    let set_x86_flags = |width| {
        connect(x86_cf, carries[&width]);
        connect(x86_of, carries[&width].cmp_ne(carries[&(width - 1)]));
        connect(x86_sf, sum[width - 1]);
        connect(x86_zf, sum[..width].cmp_eq(0u8));
    };
    #[hdl]
    let pwr_ca = wire();
    connect(pwr_ca, carries[&64]);
    #[hdl]
    let pwr_ca32 = wire();
    connect(pwr_ca32, carries[&32]);
    #[hdl]
    let pwr_ov = wire();
    connect(pwr_ov, carries[&64] ^ carries[&63]);
    #[hdl]
    let pwr_ov32 = wire();
    connect(pwr_ov32, carries[&32] ^ carries[&31]);
    #[hdl]
    let pwr_cr_lt = wire();
    connect(pwr_cr_lt, int_fp[63]);
    #[hdl]
    let pwr_cr_eq = wire();
    connect(pwr_cr_eq, int_fp.cmp_eq(0u64));
    #[hdl]
    let pwr_cr_gt = wire();
    connect(pwr_cr_gt, !pwr_cr_lt & !pwr_cr_eq);
    #[hdl]
    let pwr_so = wire();
    // TODO: SO needs to OR-in the previous value of SO
    connect(pwr_so, pwr_ov);

    #[hdl]
    match output_integer_mode {
        OutputIntegerMode::Full64 => {
            set_x86_flags(64);
            connect(int_fp, sum);
        }
        OutputIntegerMode::DupLow32 => {
            set_x86_flags(32);
            connect(
                int_fp,
                [sum.cast_to_static::<UInt<32>>(); 2]
                    .cast_to_bits()
                    .cast_to_static(),
            );
        }
        OutputIntegerMode::ZeroExt32 => {
            set_x86_flags(32);
            connect(int_fp, sum.cast_to_static::<UInt<32>>().cast_to_static());
        }
        OutputIntegerMode::SignExt32 => {
            set_x86_flags(32);
            connect(int_fp, sum.cast_to_static::<SInt<32>>().cast_to_static());
        }
        OutputIntegerMode::ZeroExt16 => {
            set_x86_flags(16);
            connect(int_fp, sum.cast_to_static::<UInt<16>>().cast_to_static());
        }
        OutputIntegerMode::SignExt16 => {
            set_x86_flags(16);
            connect(int_fp, sum.cast_to_static::<SInt<16>>().cast_to_static());
        }
        OutputIntegerMode::ZeroExt8 => {
            set_x86_flags(8);
            connect(int_fp, sum.cast_to_static::<UInt<8>>().cast_to_static());
        }
        OutputIntegerMode::SignExt8 => {
            set_x86_flags(8);
            connect(int_fp, sum.cast_to_static::<SInt<8>>().cast_to_static());
        }
    }
    #[hdl]
    let flags = wire();
    #[hdl]
    match flags_mode {
        FlagsMode::PowerISA(_) => {
            PRegFlagsPowerISA::clear_unused(flags);
            connect(PRegFlagsPowerISA::xer_ca(flags), pwr_ca);
            connect(PRegFlagsPowerISA::xer_ca32(flags), pwr_ca32);
            connect(PRegFlagsPowerISA::xer_ov(flags), pwr_ov);
            connect(PRegFlagsPowerISA::xer_ov32(flags), pwr_ov32);
            connect(PRegFlagsPowerISA::cr_lt(flags), pwr_cr_lt);
            connect(PRegFlagsPowerISA::cr_gt(flags), pwr_cr_gt);
            connect(PRegFlagsPowerISA::cr_eq(flags), pwr_cr_eq);
            connect(PRegFlagsPowerISA::so(flags), pwr_so);
        }
        FlagsMode::X86(_) => {
            PRegFlagsX86::clear_unused(flags);
            connect(PRegFlagsX86::cf(flags), x86_cf);
            connect(PRegFlagsX86::af(flags), x86_af);
            connect(PRegFlagsX86::of(flags), x86_of);
            connect(PRegFlagsX86::sf(flags), x86_sf);
            connect(PRegFlagsX86::pf(flags), x86_pf);
            connect(PRegFlagsX86::zf(flags), x86_zf);

            // this insn doesn't write DF, so it's output isn't used for reading DF
            connect(PRegFlagsX86::df(flags), false);
        }
    }
    #[hdl]
    UnitResultCompleted::<_> {
        value: #[hdl]
        PRegValue { int_fp, flags },
        extra_out: (),
    }
}

#[hdl]
fn logical(
    mop: Expr<LogicalMOp<UnitOutRegNum<DynSize>, DynSize>>,
    flags_mode: Expr<FlagsMode>,
    src_values: Expr<Array<PRegValue, { COMMON_MOP_SRC_LEN }>>,
) -> Expr<UnitResultCompleted<()>> {
    // TODO: finish
    #[hdl]
    UnitResultCompleted::<_> {
        value: PRegValue::zeroed(),
        extra_out: (),
    }
}

#[hdl_module]
pub fn alu_branch(config: &CpuConfig, unit_index: usize) {
    #[hdl]
    let cd: ClockDomain = m.input();
    #[hdl]
    let unit_to_reg_alloc: UnitToRegAlloc<
        AluBranchMOp<UnitOutRegNum<DynSize>, DynSize>,
        (),
        DynSize,
        DynSize,
        DynSize,
    > = m.output(config.unit_to_reg_alloc(
        AluBranchMOp[config.unit_out_reg_num()][config.p_reg_num_width()],
        (),
    ));
    #[hdl]
    let global_state: GlobalState = m.input();

    #[hdl]
    let unit_base = instance(unit_base(
        config,
        unit_index,
        Expr::ty(unit_to_reg_alloc).input.data.HdlSome.mop,
        (),
    ));
    connect(unit_to_reg_alloc, unit_base.unit_to_reg_alloc);
    connect(unit_base.cd, cd);
    connect(unit_base.execute_start.ready, true);
    connect(
        unit_base.execute_end,
        Expr::ty(unit_base.execute_end).HdlNone(),
    );
    #[hdl]
    if let HdlSome(execute_start) = ReadyValid::firing_data(unit_base.execute_start) {
        #[hdl]
        let ExecuteStart::<_> {
            mop,
            pc,
            src_values,
        } = execute_start;
        #[hdl]
        match mop {
            AluBranchMOp::<_, _>::AddSub(mop) => connect(
                unit_base.execute_end,
                HdlSome(
                    #[hdl]
                    ExecuteEnd::<_, _> {
                        unit_output: #[hdl]
                        UnitOutput::<_, _> {
                            which: MOpTrait::dest_reg(mop),
                            result: UnitResult[()].Completed(add_sub(
                                mop,
                                pc,
                                global_state.flags_mode,
                                src_values,
                            )),
                        },
                    },
                ),
            ),
            AluBranchMOp::<_, _>::AddSubI(mop) => connect(
                unit_base.execute_end,
                HdlSome(
                    #[hdl]
                    ExecuteEnd::<_, _> {
                        unit_output: #[hdl]
                        UnitOutput::<_, _> {
                            which: MOpTrait::dest_reg(mop),
                            result: UnitResult[()].Completed(add_sub(
                                mop,
                                pc,
                                global_state.flags_mode,
                                src_values,
                            )),
                        },
                    },
                ),
            ),
            AluBranchMOp::<_, _>::Logical(mop) => connect(
                unit_base.execute_end,
                HdlSome(
                    #[hdl]
                    ExecuteEnd::<_, _> {
                        unit_output: #[hdl]
                        UnitOutput::<_, _> {
                            which: MOpTrait::dest_reg(mop),
                            result: UnitResult[()].Completed(logical(
                                mop,
                                global_state.flags_mode,
                                src_values,
                            )),
                        },
                    },
                ),
            ),
        }
    }
}

#[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
pub struct AluBranch {
    config: Interned<CpuConfig>,
    module: Interned<Module<alu_branch>>,
}

impl AluBranch {
    pub fn new(config: &CpuConfig, unit_index: usize) -> Self {
        Self {
            config: config.intern(),
            module: alu_branch(config, unit_index),
        }
    }
}

impl UnitTrait for AluBranch {
    type Type = alu_branch;
    type ExtraOut = ();
    type MOp = AluBranchMOp<UnitOutRegNum<DynSize>, DynSize>;

    fn ty(&self) -> Self::Type {
        self.module.io_ty()
    }

    fn extra_out_ty(&self) -> Self::ExtraOut {
        ()
    }

    fn mop_ty(&self) -> Self::MOp {
        self.module.io_ty().unit_to_reg_alloc.mop_ty()
    }

    fn unit_kind(&self) -> UnitKind {
        UnitKind::AluBranch
    }

    fn extract_mop(
        &self,
        mop: Expr<UnitMOp<UnitOutRegNum<DynSize>, DynSize>>,
    ) -> Expr<HdlOption<Self::MOp>> {
        UnitMOp::alu_branch_mop(mop)
    }

    fn module(&self) -> Interned<Module<Self::Type>> {
        self.module
    }

    fn unit_to_reg_alloc(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<UnitToRegAlloc<Self::MOp, Self::ExtraOut, DynSize, DynSize, DynSize>> {
        this.unit_to_reg_alloc
    }

    fn cd(&self, this: Expr<Self::Type>) -> Expr<ClockDomain> {
        this.cd
    }

    fn global_state(&self, this: Expr<Self::Type>) -> Expr<GlobalState> {
        this.global_state
    }

    fn to_dyn(&self) -> DynUnit {
        DynUnitWrapper(*self).to_dyn()
    }
}
